############################################################################
# tools/pynuttx/nxgdb/kasan.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import argparse
import itertools

import gdb

from . import utils

# Detect which kasan is enabled by check source file name of function
kasan_set_poison = utils.get_static_symbol("kasan_set_poison")
kasan_file = kasan_set_poison.symtab.filename if kasan_set_poison else ""

CONFIG_MM_KASAN_GENERIC = "generic.c" in kasan_file
CONFIG_MM_KASAN_SW_TAGS = "sw_tags.c" in kasan_file
CONFIG_MM_KASAN_GLOBAL = utils.lookup_type("struct kasan_global_region_s") is not None
CONFIG_MM_KASAN = (
    CONFIG_MM_KASAN_GENERIC or CONFIG_MM_KASAN_SW_TAGS or CONFIG_MM_KASAN_GLOBAL
)


class KASanGeneric:
    def __init__(self, scope, begin, end, shadow, bitwidth, scale, shift=0):
        self.scope = scope
        self.begin = int(begin)
        self.end = int(end)
        self.shadow = shadow
        self.bitwidth = int(bitwidth)
        self.scale = int(scale)
        self.shift = int(shift)

    def check_addr(self, addr):
        return self.get_allocation_tag(addr)

    def contains(self, addr):
        return self.begin <= self.untag_addr(addr) <= self.end

    def get_allocation_tag(self, addr):
        distance = (addr - self.begin) // self.scale
        index = distance // self.bitwidth
        bit = distance % self.bitwidth
        return True if self.shadow[index] >> bit & 0x01 else False

    def get_logical_tag(self, addr):
        return int(addr >> self.shift)

    def untag_addr(self, addr):
        return addr


class KASanSwtags(KASanGeneric):
    def check_addr(self, addr):
        return self.get_allocation_tag(addr) != self.get_logical_tag(addr)

    def get_allocation_tag(self, addr):
        untag_addr = self.untag_addr(addr)
        distance = untag_addr - self.begin
        index = distance // self.scale
        return self.shadow[index]

    def untag_addr(self, addr):
        return addr & ~(0xFF << self.shift)


command_actions = {
    "check": lambda addr, region: print(
        f"[{region.scope}] Addr 0x{addr:X} is {'invalid' if region.check_addr(addr) else 'valid'}"
    ),
    "print-allocation-tag": lambda addr, region: print(
        f"[{region.scope}] Addr 0x{addr:X} allocation tag is: 0x{region.get_allocation_tag(addr):X}"
    ),
    "print-logical-tag": lambda addr, region: print(
        f"[{region.scope}] Addr 0x{addr:X} logical tag is: 0x{region.get_logical_tag(addr):X}"
    ),
}


class KASan(gdb.Command):

    def __init__(self):
        if CONFIG_MM_KASAN is None:
            return

        super().__init__("kasan", gdb.COMMAND_USER)

    @utils.dont_repeat_decorator
    def invoke(self, args, from_tty):
        parser = argparse.ArgumentParser(
            description="Memory Tagging Commands", add_help=False
        )
        subparsers = parser.add_subparsers(dest="command")
        subparsers.add_parser(
            "check",
            help="Validate a pointer's logical tag against the allocation tag.",
        )
        subparsers.add_parser(
            "print-allocation-tag", help="Print the allocation tag for ADDRESS."
        )
        subparsers.add_parser(
            "print-logical-tag", help="Print the logical tag from POINTER."
        )

        try:
            parsed_args, addrparser = parser.parse_known_args(gdb.string_to_argv(args))
        except SystemExit:
            return

        if not parsed_args.command:
            print(parser.format_help())
            return

        regions = {}
        address = []

        for args in addrparser:
            split = args.split(":")
            addr = utils.parse_arg(split[0])
            if len(split) == 2:
                address.extend([addr + i for i in range(0, int(split[1]))])
            else:
                address.append(addr)

        """ Common bit width and kasan alignment length in multiple modes """
        bitwidth = utils.sizeof("long") * 8
        KASAN_SHADOW_SCALE = utils.sizeof("uintptr_t")
        scale = KASAN_SHADOW_SCALE

        """ Get the array of KASan regions """
        region_count = utils.parse_and_eval("g_region_count")
        region = utils.parse_and_eval("g_region")

        self.regions: list[KASanGeneric] = []
        if CONFIG_MM_KASAN_GENERIC:
            print("KASan Mode: Generic")
            for region in utils.ArrayIterator(region, region_count):
                self.regions.append(
                    KASanGeneric(
                        "Heap",
                        region["begin"],
                        region["end"],
                        region["shadow"],
                        bitwidth,
                        scale,
                    )
                )

        if CONFIG_MM_KASAN_SW_TAGS:
            print("KASan Mode: Softtags")
            KASAN_TAG_SHIFT = 56
            shift = KASAN_TAG_SHIFT
            for region in utils.ArrayIterator(region, region_count):
                self.regions.append(
                    KASanSwtags(
                        "Heap",
                        region["begin"],
                        region["end"],
                        region["shadow"],
                        bitwidth,
                        scale,
                        shift,
                    )
                )

        if CONFIG_MM_KASAN_GLOBAL:
            print("KASan Support Checking Global Variables")
            scale = utils.parse_and_eval("g_kasan_global_align")
            global_region = utils.parse_and_eval("g_global_region")
            for index in itertools.count(0):
                if global_region[index] == 0:
                    break
                region = global_region[index].dereference()
                self.regions.append(
                    KASanGeneric(
                        "Globals",
                        region["begin"],
                        region["end"],
                        region["shadow"],
                        bitwidth,
                        scale,
                    )
                )

        for addr in address:
            for region in self.regions:
                if region.contains(addr):
                    regions[addr] = region
                    break
            if addr not in regions:
                print(f"Addr 0x{addr:X} Not in any region")

        if parsed_args.command in command_actions:
            for addr, region in regions.items():
                command_actions[parsed_args.command](addr, region)
        else:
            print(f"Unknown command: {parsed_args.command}")



############################################################################
# tools/pynuttx/nxgdb/nxcrash/busyloop.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import gdb

from .. import backtrace, utils
from ..thread import CONFIG_SCHED_CPULOAD_NONE

BUSYLOOP_THRESHOLD = 0.9 / utils.get_ncpus()


class CrashBusyloop(gdb.Command):
    """Analyse and collect the busyloop threads"""

    def __init__(self):
        if not CONFIG_SCHED_CPULOAD_NONE:
            super().__init__("crash busyloop", gdb.COMMAND_USER)

    def collect(self, tcbs):
        """Get busyloop thread"""
        collected = []
        cpuload_total = int(utils.parse_and_eval("g_cpuload_total"))

        for tcb in tcbs:
            cpuload = int(tcb["ticks"]) / cpuload_total
            if cpuload > BUSYLOOP_THRESHOLD and not utils.task_is_idle(tcb):
                collected.append((tcb, cpuload))

        return collected

    @utils.dont_repeat_decorator
    def invoke(self, arg: str, from_tty: bool) -> None:
        collected = self.collect(utils.get_tcbs())
        if not collected:
            gdb.write("No busyloop threads found.\n")
            return

        print(f"Found busyloop thread\n{'PID':<4} {'Name':<10} Load")
        for tcb, cpuload in collected:
            print(
                "{:<4} {:<10} {}".format(tcb["pid"], utils.get_task_name(tcb), cpuload)
            )

    def diagnose(self, *args, **kwargs):
        collected = self.collect(utils.get_tcbs())

        return {
            "title": "System Busyloop Detection",
            "summary": (
                f"{'No' if not collected else len(collected)} threads occur busyloop"
            ),
            "result": "fail" if collected else "pass",
            "category": utils.DiagnoseCategory.sched,
            "command": "crash busyloop",
            "thread": [
                {
                    "pid": tcb["pid"],
                    "name": utils.get_task_name(tcb),
                    "cpuload": cpuload,
                    "backtrace": backtrace.Backtrace(
                        utils.get_backtrace(int(tcb["pid"])), break_null=False
                    ),
                }
                for tcb, cpuload in collected
            ],
        }



############################################################################
# tools/pynuttx/nxgdb/debug.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import argparse

import gdb

from . import utils


class DebugPy(gdb.Command):
    """Start debugpy server, so we can debug python code from IDE like VSCode"""

    def get_argparser(self):
        parser = argparse.ArgumentParser(description=DebugPy.__doc__)
        parser.add_argument(
            "-p",
            "--port",
            default=5678,
            type=int,
            help="Server listening port",
        )
        return parser

    def __init__(self):
        debugpy = utils.import_check("debugpy", errmsg="Please pip install debugpy\n")
        if not debugpy:
            return

        self.debugpy = debugpy
        super().__init__("debugpy", gdb.COMMAND_USER)
        self.parser = self.get_argparser()

    @utils.dont_repeat_decorator
    def invoke(self, args, from_tty):
        debugpy = self.debugpy
        if debugpy.is_client_connected():
            gdb.write("debugpy is already running.\n")
            return
        try:
            args = self.parser.parse_args(gdb.string_to_argv(args))
        except SystemExit:
            return

        debugpy.listen(args.port)
        gdb.write(f"Waiting for connection at localhost:{args.port}\n")
        debugpy.wait_for_client()
        gdb.write("Debugger connected.\n")



############################################################################
# tools/pynuttx/nxgdbmcp/src/gmcp/tools/controlflow.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import re
from typing import Optional

from mcp.server.fastmcp import Context

from ..context import get_session
from ..utils import _exec_command, error_handler


def _format_breakpoint_result(
    location: str, condition: Optional[str] = None, output: str = ""
) -> str:
    msg = f"Breakpoint set at: {location}"
    if condition:
        msg += f" with condition: {condition}"
    msg += "\n\nOutput:\n" + output
    return msg


@error_handler
async def _set_breakpoint(
    ctx: Context, session_id: str, location: str, condition: Optional[str] = None
):
    session = get_session(ctx, session_id)
    output = await session.execute_command(f"break {location}")
    if condition:
        match = re.search(r"Breakpoint (\d+)", output)
        if match:
            bp_num = match.group(1)
            condition_output = await session.execute_command(
                f"condition {bp_num} {condition}"
            )
            output += f"\n{condition_output}"
    return _format_breakpoint_result(location, condition, output)


@error_handler
async def _step(ctx: Context, session_id: str, instructions: bool = False) -> str:
    session = get_session(ctx, session_id)
    command = "stepi" if instructions else "step"
    output = await session.execute_command(command)
    return (
        f"Stepped over {'instruction' if instructions else 'function call'}\n\nOutput:\n"
        f"{output}"
    )


@error_handler
async def _next(ctx: Context, session_id: str, instructions: bool = False) -> str:
    session = get_session(ctx, session_id)
    command = "nexti" if instructions else "next"
    output = await session.execute_command(command)
    return f"Stepped over {'instruction' if instructions else 'function call'}\n\nOutput:\n{output}"


@error_handler
async def _backtrace(
    ctx: Context, session_id: str, full: bool = False, limit: Optional[int] = None
) -> str:
    session = get_session(ctx, session_id)

    # Build backtrace command with options
    command = "bt"
    if full:
        command += " full"
    if limit is not None:
        command += f" {limit}"

    output = await session.execute_command(command)
    limit_str = f" (limit: {limit})" if limit else ""
    return f"Backtrace{' (full)' if full else ''}{limit_str}:\n\n{output}"


@error_handler
async def _process_info(ctx: Context, session_id: str) -> str:
    session = get_session(ctx, session_id)
    output = await session.execute_command("info program")
    pid_output = await session.execute_command("info proc")
    return f"Process information:\n\n{output}\n\nDetails:\n{pid_output}"


def register_control_flow_tools(gdb_mcp):
    @gdb_mcp.tool()
    async def gdb_set_breakpoint(
        ctx: Context, session_id: str, location: str, condition: Optional[str] = None
    ) -> str:
        """Set a breakpoint at the specified location.

        Args:
            session_id: The GDB session identifier
            location: A string specifying where to set the breakpoint
                (e.g., "main", "file.c:123", "*0x12345678")
            condition: Optional string containing a conditional expression
                (e.g., "x > 10", "ptr != NULL")
        """
        return await _set_breakpoint(ctx, session_id, location, condition)

    @gdb_mcp.tool()
    async def gdb_continue(ctx: Context, session_id: str) -> str:
        """Continue program execution until the next breakpoint or program termination.

        Args:
            session_id: The GDB session identifier
        """
        return await _exec_command(ctx, session_id, "continue")

    @gdb_mcp.tool()
    async def gdb_step(
        ctx: Context, session_id: str, instructions: bool = False
    ) -> str:
        """Step into the next line of code, entering function calls.

        Args:
            session_id: The GDB session identifier
            instructions: If True, step one machine instruction; if False, step one source line
        """
        return await _step(ctx, session_id, instructions)

    @gdb_mcp.tool()
    async def gdb_next(
        ctx: Context, session_id: str, instructions: bool = False
    ) -> str:
        """Step over the next line of code, not entering function calls.

        Args:
            session_id: The GDB session identifier
            instructions: If True, step over one machine instruction; if False, step over one source line
        """
        return await _next(ctx, session_id, instructions)

    @gdb_mcp.tool()
    async def gdb_finish(ctx: Context, session_id: str) -> str:
        """Execute until the current function returns.

        Args:
            session_id: The GDB session identifier
        """
        return await _exec_command(ctx, session_id, "finish")

    @gdb_mcp.tool()
    async def gdb_backtrace(
        ctx: Context, session_id: str, full: bool = False, limit: Optional[int] = None
    ) -> str:
        """Display the call stack backtrace.

        Args:
            session_id: The GDB session identifier
            full: If True, display local variables and arguments for each frame
            limit: Optional integer to limit the number of frames displayed
        """
        return await _backtrace(ctx, session_id, full, limit)

    @gdb_mcp.tool()
    async def gdb_breakpoint_list(ctx: Context, session_id: str) -> str:
        """List all currently set breakpoints.

        Args:
            session_id: The GDB session identifier
        """
        return await _exec_command(ctx, session_id, "info breakpoints")

    @gdb_mcp.tool()
    async def gdb_breakpoint_delete(
        ctx: Context, session_id: str, breakpoint_id: int
    ) -> str:
        """Delete a specific breakpoint by its ID.

        Args:
            session_id: The GDB session identifier
            breakpoint_id: The numeric ID of the breakpoint to delete (integer, not a string)
        """
        return await _exec_command(
            ctx, session_id, f"delete breakpoints {breakpoint_id}"
        )

    @gdb_mcp.tool()
    async def gdb_thread_list(ctx: Context, session_id: str) -> str:
        """List all threads in the current host process.

        Args:
            session_id: The GDB session identifier
        """
        return await _exec_command(ctx, session_id, "info threads")

    @gdb_mcp.tool()
    async def gdb_thread_select(ctx: Context, session_id: str, thread_id: int) -> str:
        """Select a specific thread by GDB thread ID.

        Args:
            session_id: The GDB session identifier
            thread_id: The GDB thread ID (integer, not a string). Use gdb_thread_list to see available IDs.
        """
        return await _exec_command(ctx, session_id, f"thread {thread_id}")

    @gdb_mcp.tool()
    async def gdb_process_info(ctx: Context, session_id: str) -> str:
        """Get information about the current process being debugged.

        Args:
            session_id: The GDB session identifier
        """
        return await _process_info(ctx, session_id)


def register_nxthread_tools(gdb_mcp):
    @gdb_mcp.tool()
    async def nxgdb_list_thread(ctx: Context, session_id: str) -> str:
        """List all NuttX threads in the current system.

        Executes the 'info nxthread' command to retrieve thread information using the
        NuttX thread model, which provides more detailed thread data than standard GDB
        thread listings.

        Args:
            ctx (Context): The context object containing runtime/session information.
            session_id (str): Identifier for the GDB debugging session.

        Returns:
            str: Detailed thread information from the NuttX thread model including
                 thread IDs, states, priorities, and stack information.
        """
        return await _exec_command(ctx, session_id, "info nxthreads")

    @gdb_mcp.tool()
    async def nxgdb_thread_apply(
        ctx: Context, session_id: str, ids: str, cmd: str
    ) -> str:
        """Apply a GDB command to some NuttX threads using 'nthread apply'.

        Executes the 'nthread apply <all|id list> <cmd>' command, where:
        - 'all' applies the command to all NuttX threads
        - 'id list' applies the command to specific thread IDs (comma or space separated)
        - 'cmd' is the GDB command to execute (e.g., 'bt')

        Args:
            session_id: Identifier for the GDB debugging session
            ids: A string that is either "all" or a list of NuttX thread ID Hashes (e.g., "1,2,3" or "1 2 3")
            cmd: A string containing the GDB command to execute for each thread (e.g., "bt", "info registers")

        Returns:
            str: The output from the 'nthread apply' command for the specified threads.
        """
        return await _exec_command(ctx, session_id, f"nthread apply {ids} {cmd}")

    @gdb_mcp.tool()
    async def nxgdb_thread_select(ctx: Context, session_id: str, thread_id: int) -> str:
        """Select a specific NuttX thread by pid.

        The 'nxthread <pid>' command switches to the thread with the given NuttX thread model pid
        by restoring the registers for that thread context.

        Args:
            session_id: The GDB session identifier
            thread_id: The NuttX thread PID (integer, not a string). Use nxgdb_list_thread to see available PIDs.
        """
        return await _exec_command(ctx, session_id, f"nxthread {thread_id}")



############################################################################
# tools/pynuttx/nxgdb/diagnose.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import argparse
from gc import get_referrers

import gdb

from . import autocompeletion, utils


class DiagnosePrefix(gdb.Command):
    """Diagnostic related commands."""

    def __init__(self):
        super().__init__("diagnose", gdb.COMMAND_USER, prefix=True)


@autocompeletion.complete
class DiagnoseReport(gdb.Command):
    """Run diagnostics to generate reports."""

    def get_argparser(self):
        parser = argparse.ArgumentParser(description=self.__doc__)
        parser.add_argument(
            "-o",
            "--output",
            type=str,
            metavar="file",
            help="report output file name",
        )
        parser.add_argument(
            "-c",
            "--command",
            nargs="+",
            help="only run specified commands (case-insensitive)",
        )
        parser.add_argument(
            "-l", "--list", action="store_true", help="list available commands"
        )
        return parser

    def __init__(self):
        super().__init__("diagnose report", gdb.COMMAND_USER)
        self.parser = self.get_argparser()

    @utils.dont_repeat_decorator
    def invoke(self, args, from_tty):
        try:
            args = self.parser.parse_args(gdb.string_to_argv(args))
        except SystemExit:
            return

        reportfile = (
            args.output
            if args.output
            else gdb.objfiles()[0].filename + ".diagnostics.json"
        )

        modules = utils.gather_modules()
        modules.remove("prefix")

        commands = utils.gather_gdbcommands(modules=modules)

        registered_command_types = {
            obj.__class__.__name__.lower()
            for cls in gdb.Command.__subclasses__()
            for obj in get_referrers(cls)
            if isinstance(obj, cls) and not isinstance(obj, type)
        }

        only_set = set(name.lower() for name in args.command) if args.command else None

        results = []
        for clz in commands:
            if hasattr(clz, "diagnose"):
                command = clz()
                name = clz.__name__.lower()
                if name not in registered_command_types:
                    continue

                if args.list:
                    gdb.write(f"{name}\n")
                    continue

                if only_set:
                    if name not in only_set:
                        continue

                gdb.write(f"Run command: {name}\n")
                try:
                    result = command.diagnose()
                except Exception as e:
                    result = {
                        "title": f"Command {name} failed",
                        "summary": "Command execution failed",
                        "result": "info",
                        "command": name,
                        "message": str(e),
                    }

                    gdb.write(f"Failed: {e}\n")

                result.setdefault("command", name)
                results.append(result)

        if args.list:
            return

        gdb.write(f"Write report to {reportfile}\n")
        with open(reportfile, "w") as f:
            f.write(utils.jsonify(results, indent=4))



############################################################################
# tools/pynuttx/nxstub/__init__.py
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

__version__ = "0.0.1"

import argparse
import logging
import multiprocessing
import os
import re
import signal
import socket
import subprocess
import traceback
from typing import List

from nxreg.register import Registers, g_reg_table

from . import utils
from .gdbstub import GDBStub, Target
from .proxy import TargetProxy
from .target import RawMemory

# Default GDB init command, note it should start with a space
DEFAULT_GDB_INIT_CMD = " -ex 'bt full' -ex 'info reg' -ex 'display /40i $pc-40'"


def parse_log(elf, arch, logfile):
    memories: List[List[RawMemory]] = []
    registers: List[Registers] = []
    if not logfile:
        return Registers(elf, arch), memories

    def is_register_dump(line):
        return len(line.split("up_dump_register: ")) == 2

    def parse_register(regs, line):
        line = line.split("up_dump_register: ")
        if not (find_res := re.findall(r"([\w_]+): *([0-9xa-fXA-F]+)", line[1])):
            return False

        for name, value in find_res:
            name = name.lower()
            value = int(value, 16)
            try:
                regs.set(value, name=name)
            except KeyError:
                logging.warning(f"Ignore register {name}:{value}")

        return True

    def parse_stack(line):
        result = re.match(
            r".*stack_dump: (?P<ADDR>[0-9a-fxA-FX]+): (?P<VALS>( ?\w+)+)", line
        )
        if result is None:
            return None

        results = result.groupdict()

        addr = int(results["ADDR"], 16)
        data = b""
        for val in results["VALS"].split():
            # For little endian, the hex bytes should be reversed
            data += bytes.fromhex(val)[::-1]
        return RawMemory(addr, data)

    registerdump = []
    tempmemories = []
    parsing_regs = False
    lines = []
    with open(logfile, "r", errors="ignore") as f:
        for line in f:
            line = line.strip()
            if is_register_dump(line):
                if not parsing_regs:
                    parsing_regs = True
                    if lines:  # Started
                        registers.append(tmpregisters)  # noqa
                        memories.append(tempmemories)  # noqa
                        registerdump.append(lines)
                    lines = []
                    tempmemories = []
                    tmpregisters = Registers(elf, arch)

                lines.append(line)
                parse_register(tmpregisters, line)
            elif memory := parse_stack(line):
                parsing_regs = False
                tempmemories.append(memory)

        if lines:
            registers.append(tmpregisters)
            memories.append(tempmemories)
            registerdump.append(lines)

    if len(memories) > 1:
        for i, line in enumerate(registerdump):
            print(f"[{i}]: {'    '.join(line[:2])}")

        choice = input("Multiple dump found, please choose one:\n").strip() or 0
    else:
        choice = 0

    return registers[int(choice)], memories[int(choice)]


def auto_parse_dump(args):
    """
    Automatically parse the dump file, which could be a crash log, memory dump or core dump.
    Store the parsed result directly to args, so the remaining logic keeps unchanged.
    """
    dump = args.dump
    if not dump:
        return

    if args.rawfile or args.log or args.core:
        raise ValueError("Error: 'dump' cannot be used with rawfile, log, or core.")

    if os.path.isdir(dump):
        # We suppose only memory dump could be a directory
        print(f"Input is raw memory dump: {dump}")
        args.rawfile = [dump]  # rawfile must be a list
        return

    def is_core_file(file):
        # check elf header, check magic and elf type is CORE
        with open(dump, "rb") as f:
            magic = f.read(4)
            if magic != b"\x7fELF":
                return False

            f.seek(0x10)
            elf_type = int.from_bytes(f.read(2), "little")
            return elf_type == 4

    # Check if the dump file is a crash log, memory dump or core dump
    if dump.endswith(".log"):
        print(f"Input is crash log: {dump}")
        args.log = dump
    elif dump.endswith(".bin"):
        print(f"Input is raw memory dump: {dump}")
        args.rawfile = [dump]
    elif is_core_file(dump):
        print(f"Input is core dump: {dump}")
        args.core = dump
    else:
        raise ValueError(f"Unknown dump file type: {dump}")


def gdbstub_start(args):
    memories = []
    registers = None

    # Parse args.dump to normal parameter if exist
    auto_parse_dump(args)

    for name in args.rawfile or []:

        def get_address(filename: str):
            """
            Get memory dump address from file name from below formats
            memdump.bin:0x123456
            abc/0x123456.bin
            0x123456.bin
            abc/123456.bin
            123456.bin
            """
            try:
                if ":" in filename:
                    return int(filename.split(":")[1], 16)
                else:
                    return int(filename.split("/")[-1].split(".")[0], 0)
            except ValueError:
                return None

        if os.path.isdir(name):
            for f in os.listdir(name):
                if (address := get_address(f)) is None:
                    print(f"Ignore file {os.path.join(name, f)}")
                else:
                    with open(os.path.join(name, f), "rb") as f:
                        memories.append(RawMemory(address, f.read()))
                        print(f"Add memory dump: {memories[-1]}")
        else:
            address = get_address(name)
            with open(name.split(":")[0], "rb") as f:
                memories.append(RawMemory(address, f.read()))
                print(f"Add memory dump: {memories[-1]}")

    elf = utils.LiefELF(args.elffile)
    registers, mem = parse_log(elf, args.arch, args.log)
    memories.extend(mem)

    core = utils.LiefELF(args.core) if args.core else None

    memremap = []
    if args.remap:
        for remap in args.remap:
            fromaddr, toaddr, length = map(lambda x: int(x, 16), remap.split(","))
            memremap.append((fromaddr, toaddr, length))

    if args.proxy is not None:
        print(f"Try proxying localhost:{args.proxy}...")
        target = TargetProxy(elf, args.arch, args.proxy)
    else:
        target = Target(elf, args.arch, registers, memories, memremap, core)

    stub = GDBStub(target=target, port=args.port, proxymode=args.proxy is not None)

    print(f"Start GDB server on port {args.port}...")
    stub.run()
    print("GDB server exited.")


def gdb_start(args):
    def gdb_run(cmd):
        try:
            subprocess.run(cmd, shell=True)
        except KeyboardInterrupt:
            pass

    gdb_init_cmd = args.init_cmd or DEFAULT_GDB_INIT_CMD
    gdb_exec = args.gdb or "gdb-multiarch"

    timeout = f"set tcp connect-timeout {args.timeout}" if args.timeout else ""
    if timeout:
        print(f"Set GDB timeout: {args.timeout}")

    gdb_cmd = (
        f"{gdb_exec} {args.elffile} -ex '{timeout}' -ex 'target remote localhost:{args.port}' "
        f"{gdb_init_cmd}"
    )
    print(f"Start GDB session: {gdb_cmd}")
    return multiprocessing.Process(target=gdb_run, args=(gdb_cmd,)).start()


def get_argparser():
    parser = argparse.ArgumentParser(
        prog="nxstub",
        description=f"nxstub v{__version__} - NuttX GDB server based on crash log, core dump or memory dump.",
    )
    parser.add_argument(
        "-a",
        "--arch",
        type=str,
        choices=g_reg_table.keys(),
        required=True,
        help="The architecture of the target.",
    )
    parser.add_argument(
        "-e",
        "--elffile",
        type=str,
        metavar="file",
        required=True,
        help="The elf file.",
    )
    parser.add_argument(
        "-p",
        "--port",
        type=int,
        default=1234,
        help="The GDB server port.",
    )
    parser.add_argument(
        "--proxy",
        type=int,
        default=None,
        help="The original GDB server port for proxy.",
    )
    parser.add_argument(
        "-r",
        "--rawfile",
        type=str,
        metavar="file",
        nargs="*",
        help="The memory dump file, in format of 'memdump1.bin:address1 memdump2.bin:address2'.",
    )
    parser.add_argument(
        "-c",
        "--core",
        type=str,
        metavar="file",
        help="The core dump file.",
    )
    parser.add_argument(
        "--remap",
        type=str,
        nargs="*",
        help="Remap the memory to another address, argument in format of 'from,to,length'.",
    )
    parser.add_argument(
        "-l",
        "--log",
        type=str,
        metavar="file",
        help="The crash dump log file.",
    )
    parser.add_argument(
        "-d",
        "--debug",
        action="store_true",
        help="Show debug messages.",
    )
    parser.add_argument(
        "-g",
        "--gdb",
        help="Optional path to GDB executable, once specified, will automatically start GDB session.",
        type=str,
        metavar="file",
    )
    parser.add_argument(
        "--timeout",
        type=str,
        help="Timeout in seconds for GDB to wait for gdbserver to startup. Use 'unlimited' or integer value.",
    )
    parser.add_argument(
        "-i",
        "--init-cmd",
        type=str,
        help=f'Optional custom GDB init command when GDB launches, default: "{DEFAULT_GDB_INIT_CMD}".',
    )
    parser.add_argument(
        "dump",
        type=str,
        metavar="file",
        nargs="?",
        default=None,
        help="Optional dump file that could be crash log, memory dump or core dump, automatically parsed.",
    )
    return parser


def parse_args(args=None):
    return get_argparser().parse_args(args)


def get_unused_port():
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.bind(("", 0))
        return s.getsockname()[1]


def is_port_in_use(port: int) -> bool:
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        try:
            s.bind(("127.0.0.1", port))
        except OSError:
            return True
        return False


def main(args):
    if args.debug:
        logging.basicConfig(
            format="%(asctime)s:%(levelname)s:%(name)s:%(message)s",
        )
        logging.getLogger(__name__).setLevel(logging.DEBUG)
        logging.getLogger("nxreg").setLevel(logging.DEBUG)

    if is_port_in_use(args.port):
        print(f"Port {args.port} is already in use, try to use another port.")
        args.port = get_unused_port()
        print(f"Use port {args.port} instead.")

    gdb = None
    if args.gdb:
        gdb = gdb_start(args)
        # Ignore the Ctrl+C signal
        signal.signal(signal.SIGINT, signal.SIG_IGN)

    try:
        gdbstub_start(args)
    except Exception as e:
        print(f"GDBStub error: {e}\n {traceback.format_exc() if args.debug else ''}")

    if gdb:
        print("Stop GDB session...")
        gdb.kill()
        gdb.join()



############################################################################
# tools/gdb/nuttxgdb/memclassify.py
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################
import argparse
import math
import os
import sys

import gdb

from . import autocompeletion, backtrace, memdump, mm, utils


class Classifier:
    def __init__(self, info, category_name=""):
        self.stats = []
        self.judgers = []
        self.class_name = category_name

        for key, sub_info in info.items():
            ignore = False
            if callable(sub_info):
                self.judgers.append(sub_info)
                sub_classifier = None
            else:  # sub_info is a dict
                if "ignore" in sub_info:
                    ignore = sub_info["ignore"]
                self.judgers.append(sub_info["judger"])
                sub_classifier = (
                    Classifier(
                        sub_info["subcategories"], "_".join([category_name, key])
                    )
                    if "subcategories" in sub_info
                    else None
                )

            self.stats.append(MemoryCategory(key, ignore, sub_classifier))

    def __call__(self, mem_blocks):
        for mb in mem_blocks:
            cat = next(
                (idx for idx, method in enumerate(self.judgers) if method(mb)),
                None,
            )

            if cat is None:
                cat = len(self.stats)
                self.stats.append(MemoryCategory(category="unknown"))
                self.judgers.append(lambda x: True)
            self.stats[cat].append(mb)
        for stat in self.stats:
            stat.classify()
        return self.stats


class MemBlock:
    def __init__(self):
        pass

    def total_size(self):
        pass

    def total_size_without_overhead(self):
        pass

    def count(self):
        pass

    def block_size(self):
        pass

    def overhead_size(self):
        pass

    def pid(self):
        pass

    def backtrace(self):
        pass

    def print(self, f=sys.stdout):
        def pprint(*args):
            print(*args, file=f)

        pprint(
            f"pid: {self.pid()}, size: {self.total_size()} = {self.block_size()} x {self.count()}"
        )
        pprint(
            f"size: {self.total_size_without_overhead()} = {self.block_size() - self.overhead_size()} x {self.count()}"
        )

        for name, pos in self.backtrace():
            pprint(f"{name:<50} {pos}")
        pprint("")


class MemoryCategory:
    WIDTH = 88

    def __init__(self, category="total", ignore=False, classifier=None):
        self.category = category
        self.mem_blocks = []
        self.classifier = classifier
        self.children = []
        self.ignore = ignore

    def append(self, mem_block):
        self.mem_blocks.append(mem_block)

    def extend(self, mem_blocks):
        self.mem_blocks.extend(mem_blocks)

    def classify(self):
        if self.classifier is None:
            return
        self.children = self.classifier(self.mem_blocks)

    def summarize(self):
        for child in self.children:
            child.summarize()

        self.total_mem_size = sum(mb.total_size() for mb in self.mem_blocks)
        self.no_overhead_size = sum(
            mb.total_size_without_overhead() for mb in self.mem_blocks
        )
        self.mem_block_unique_cnt = len(self.mem_blocks)
        self.mem_block_cnt = sum(mb.count() for mb in self.mem_blocks)

    def print_statistics(self, categorys=[]):

        if not self.children:  # No children means no categories, just print nothing
            return

        categorys.append(self.category)

        def formatter(
            name, total_mem_size, total_mem_cnt1, total_mem_cnt2, mem_rm_bt=0
        ):
            print(
                f"{name:<15}| {total_mem_size:<15}| {total_mem_cnt1:<15}| {total_mem_cnt2:<15}| {mem_rm_bt:<15}"
            )

        title = f"{'.'.join(categorys)} Mem Statistics".center(self.WIDTH, "-")
        print(title)
        formatter(
            "category",
            "total mem size",
            "memblk uniq cnt",
            "memblk tot cnt",
            "mem without overhead",
        )
        print("-" * self.WIDTH)
        for child in self.children:
            formatter(
                child.category,
                child.total_mem_size,
                child.mem_block_unique_cnt,
                child.mem_block_cnt,
                child.no_overhead_size,
            )

        formatter(
            "total",
            self.total_mem_size,
            self.mem_block_unique_cnt,
            self.mem_block_cnt,
            self.no_overhead_size,
        )
        print("-" * self.WIDTH)
        for child in self.children:
            child.print_statistics()
        categorys.pop()

    def collect_piedata(self, title_path=[]):
        title_path.append(self.category)
        res = []
        if self.children:
            res.append(
                (
                    ".".join(title_path),
                    [child.category for child in self.children if not child.ignore],
                    [
                        child.no_overhead_size
                        for child in self.children
                        if not child.ignore
                    ],
                )
            )
            for child in self.children:
                res.extend(child.collect_piedata(title_path))
        title_path.pop()
        return res

    def dump_category(self, output_dir, parentCat=None):
        os.makedirs(output_dir, exist_ok=True)

        for child in self.children:
            arr = sorted(child.mem_blocks, key=lambda x: x.total_size(), reverse=True)
            outputfile = (
                parentCat + f"_{child.category}" if parentCat else child.category
            )
            with open(os.path.join(output_dir, outputfile + ".bt"), "w") as f:
                print(f"{child.category}, count={len(child.mem_blocks)}", file=f)
                for mb in arr:
                    mb.print(f)
            child.dump_category(output_dir, outputfile)


def draw_pie(stat):
    datasets = stat.collect_piedata()
    plt = utils.import_check(
        "matplotlib.pyplot", errmsg="Please pip install matplotlib\n"
    )
    if plt is None:
        print("matplotlib is not installed")
        return

    num_plots = len(datasets)

    def fact_num(n):
        h = math.floor(math.sqrt(n))
        min_dlt = n
        res = []
        for a in range(1, h + 1):
            b = math.ceil(n / a)
            dlt = b - a
            if dlt < min_dlt:
                res = [a, b]
        return res

    row, col = fact_num(num_plots)
    fig, axs = plt.subplots(row, col, figsize=(12, 6), subplot_kw=dict(aspect="equal"))
    if num_plots == 1:
        axs = [axs]

    for i, (title, labels, sizes) in enumerate(datasets):
        temp = [x for x in filter(lambda x: x[1], zip(labels, sizes))]
        if len(temp) > 0:
            labels, sizes = zip(*temp)
        else:
            continue
        x, y = i // col, i % col
        axs[x, y].pie(sizes, labels=labels, autopct="%1.1f%%")
        axs[x, y].set_title(title)
    plt.tight_layout()
    plt.show()


class MemBlockCoredump(MemBlock):
    def __init__(self, node, cnt):
        super().__init__()
        self.mem_node = node
        self.cnt = cnt  # deal with call_stack
        self.call_stack = []
        for addr, func, file, line in backtrace.Backtrace(node.backtrace).backtrace:
            func = func.strip("<>")
            if func.find("+"):  # remove the offset affter '+'
                func = func[: func.find("+")]
            self.call_stack.append((func, f"{file}:{line}"))

    def total_size(self):
        return self.mem_node.nodesize * self.cnt

    def total_size_without_overhead(self):
        return (self.mem_node.nodesize - self.mem_node.overhead) * self.cnt

    def count(self):
        return self.cnt

    def block_size(self):
        return self.mem_node.nodesize

    def overhead_size(self):
        return self.mem_node.overhead

    def pid(self):
        return self.mem_node.pid

    def backtrace(self):
        return self.call_stack


@autocompeletion.complete
class MMClassify(gdb.Command):
    """classify memory by callstack"""

    def get_argparser(self):
        parser = argparse.ArgumentParser(description="Memory Classify")
        parser.add_argument(
            "-o",
            "--output-dir",
            metavar="file",
            default="memclass.output",
            help="Specify the directory to save the the call stack files after categorization",
        )

        parser.add_argument(
            "-p", "--pid", type=int, default=None, help="Thread PID, -1 for mempool"
        )

        parser.add_argument(
            "--pids", nargs="+", type=int, default=[], help="List of pids"
        )

        parser.add_argument(
            "-c",
            "--classifier-file",
            metavar="file",
            default="default",
            help="Specify the config file. Default is 'script_path/default.py'",
        )

        parser.add_argument(
            "-l",
            "--log",
            default=None,
            metavar="file",
            help="Specify the memdump log file.",
        )
        return parser

    def __init__(self):
        super().__init__("mm classify", gdb.COMMAND_USER)
        utils.alias("memclassify", "mm classify")
        self.parser = self.get_argparser()

    def parse_args(self, argv):
        try:
            args = self.parser.parse_args(argv)
        except SystemExit:
            return False

        return args

    @utils.dont_repeat_decorator
    def invoke(self, arg: str, from_tty: bool) -> None:
        if not (args := self.parse_args(gdb.string_to_argv(arg))):
            print("memoryclassify: parse args error")
            return

        def import_classify_config(classifier_file):
            classifier_dir = os.path.join(os.path.dirname(__file__), "memclassifier")
            sys.path.append(classifier_dir)
            return utils.import_reload(
                classifier_file,
                errmsg=f"Please provide {classifier_file}.py in {classifier_dir}\n",
            )

        classify_config = import_classify_config(args.classifier_file)
        if not classify_config:
            return

        memblocks = []
        if args.log:  # get mem block from memdump log
            memblocks.extend(
                MemBlockCoredump(node, 1)
                for node in memdump.parse_memdump_log(args.log)
            )
        else:
            if not mm.MM_RECORD_STACK_DEPTH < 8:
                print("memoryclassify: no backtrace")
                return
            if args.pid is not None:
                args.pids.append(args.pid)
            for pid in args.pids:
                filters = {
                    "pid": pid,
                    "nodesize": None,
                    "used": None,
                    "free": None,
                    "seqmin": None,
                    "seqmax": None,
                    "orphan": None,
                    "no_heap": None,
                    "no_pool": None,
                    "no_pid": None,
                }

                memblocks.extend(
                    MemBlockCoredump(node, len(nodes))
                    for node, nodes in memdump.group_nodes(
                        memdump.dump_nodes(mm.get_heaps(), filters)
                    ).items()
                )

        stat = MemoryCategory(
            "total", False, Classifier(getattr(classify_config, "categories"))
        )
        stat.extend(memblocks)
        stat.classify()
        stat.summarize()
        stat.print_statistics()
        stat.dump_category(args.output_dir)
        draw_pie(stat)



############################################################################
# tools/pynuttx/nxgdb/fs/__init__.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

__all__ = [
    "foreach_inode",
    "foreach_file",
    "get_fstype",
    "get_file",
    "fstype_filter",
    "inode_gettype",
    "get_inode_name",
    "inode_getpath",
    "ForeachInode",
    "Fdinfo",
    "Mount",
    "InfoShmfs",
    "InfoRomfs",
    "InfoLittlefs",
    "InfoYaffs",
    "InodeType",
    "Inode",
]

from .command import (
    Fdinfo,
    ForeachInode,
    InfoLittlefs,
    InfoRomfs,
    InfoShmfs,
    InfoYaffs,
    Mount,
)
from .fs import foreach_file, get_file
from .inode import (
    Inode,
    InodeType,
    foreach_inode,
    fstype_filter,
    get_fstype,
    get_inode_name,
    inode_getpath,
    inode_gettype,
)



############################################################################
# tools/pynuttx/nxgdb/armspunwind.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

"""
ARM Stack-based Backtrace Unwinder for GDB

This module provides stack-based backtrace unwinding for ARM architecture
without relying on unwind tables. It analyzes branch instructions (BL/BLX)
in the code to reconstruct call stacks from stack memory.

Usage:
    from arm_sp_unwind import ArmSpUnwind

    # Initialize unwinder
    unwinder = ArmSpUnwind()

    # Set custom code regions if needed
    unwinder.set_code_regions(
        [(0x8000000, 0x8100000), (0x20000000, 0x20010000)]
    )

    # Unwind from TCB
    tcb = gdb.parse_and_eval("tcb_ptr")
    addresses = unwinder.unwind_from_tcb(tcb, max_frames=10)

    # Unwind from stack memory range
    addresses = unwinder.unwind_from_stack(
        sp=0x20001000, top=0x20002000, max_frames=10
    )

    # Use with Backtrace class for pretty printing
    from backtrace import Backtrace
    backtrace = Backtrace(addresses)
    print(backtrace)
"""

import argparse
import re
from typing import List, Optional, Tuple

import gdb

from . import autocompeletion, utils
from .backtrace import Backtrace

# ARM Thumb instruction masks and opcodes
# BLX instruction: 0b010001111x000000
IMASK_T_BLX = 0xFF80
IOP_T_BLX = 0x4780

# BL instruction (high halfword): 0b11110xxxxxxxxxxx
IMASK_T_BL = 0xF800
IOP_T_BL = 0xF000

# BL instruction (low halfword): 0b11x1xxxxxxxxxxxx
IMASK_T_BL_LOW = 0xD000
IOP_T_BL_LOW = 0xD000


def parse_code_regions() -> List[Tuple[int, int]]:
    """Parse code regions from GDB target sections

    Parse output of 'maint info target-sections' to find all CODE
    sections. Returns list of (start, end) tuples for code regions.
    """

    # Execute GDB command to get target sections
    output = gdb.execute("maint info target-sections", to_string=True)

    regions = []
    # Pattern to match lines with CODE flag
    # Look for hex addresses around -> on lines containing CODE
    pattern = r"0x([0-9a-fA-F]+)->0x([0-9a-fA-F]+).*\bCODE\b"

    for line in output.split("\n"):
        match = re.search(pattern, line)
        if match:
            start_addr = int(match.group(1), 16)
            end_addr = int(match.group(2), 16)

            # Only add valid regions (start < end)
            if start_addr < end_addr:
                regions.append((start_addr, end_addr))

    if regions:
        # Sort regions by start address and merge
        # overlapping/adjacent ones
        regions.sort(key=lambda x: x[0])
        merged = []
        for start, end in regions:
            if merged and start <= merged[-1][1]:
                # Overlapping or adjacent, merge them
                merged[-1] = (
                    merged[-1][0],
                    max(merged[-1][1], end),
                )
            else:
                merged.append((start, end))

        return merged

    return []


class ArmSpUnwind:
    """
    ARM Stack-based Backtrace Unwinder

    This class provides functionality to unwind ARM call stacks by analyzing
    branch instructions (BL/BLX) stored on the stack, without requiring
    unwind tables or frame pointers.
    """

    def __init__(self):
        """Initialize the ARM stack unwinder"""
        self.code_regions = parse_code_regions()

    def set_code_regions(self, regions: List[Tuple[int, int]]):
        """
        Set custom code regions for backtrace validation

        Args:
            regions: List of (start, end) tuples defining valid code
                     address ranges.
                     Example: [(0x8000000, 0x8100000),
                               (0x20000000, 0x20010000)]
        """
        self.code_regions = regions

    def in_code_region(self, pc: int) -> bool:
        """
        Check if program counter is within valid code regions

        Args:
            pc: Program counter address to check

        Returns:
            True if PC is within a valid code region, False otherwise
        """
        if not self.code_regions:
            # If no regions defined, try to validate using symbol table
            try:
                # Try to find symbol at this address
                block = gdb.block_for_pc(pc)
                return block is not None
            except (gdb.error, Exception):
                return False

        for start, end in self.code_regions:
            if start <= pc < end:
                return True

        return False

    def is_thumb_blx_instruction(self, ins16: int) -> bool:
        """
        Check if 16-bit value is a Thumb BLX instruction

        Args:
            ins16: 16-bit instruction value

        Returns:
            True if instruction is BLX, False otherwise
        """
        return (ins16 & IMASK_T_BLX) == IOP_T_BLX

    def is_thumb_bl_instruction(self, ins16_high: int, ins16_low: int) -> bool:
        """
        Check if two 16-bit values form a Thumb BL instruction

        Thumb BL is a 32-bit instruction split into two 16-bit halfwords:
        - High halfword: 0b11110xxxxxxxxxxx (0xF000-0xF7FF or 0xF800-0xFFFF)
        - Low halfword:  0b11x1xxxxxxxxxxxx (0xD000-0xDFFF or 0xF000-0xFFFF)

        Args:
            ins16_high: High 16-bit halfword of instruction
            ins16_low: Low 16-bit halfword of instruction

        Returns:
            True if instruction is BL, False otherwise
        """
        return (ins16_high & IMASK_T_BL) == IOP_T_BL and (
            ins16_low & IMASK_T_BL_LOW
        ) == IOP_T_BL_LOW

    def read_word(self, addr: int, halfword: bool = False) -> Optional[int]:
        """
        Read a word from memory

        Args:
            addr: Memory address to read from
            halfword: If True, read 16-bit halfword; if False, read 32-bit word

        Returns:
            Word value or None if read fails
        """
        try:
            inferior = gdb.selected_inferior()
            size = 2 if halfword else 4
            mem = inferior.read_memory(addr, size)
            # ARM is typically little-endian
            return int.from_bytes(mem, byteorder="little")
        except (gdb.MemoryError, Exception):
            return None

    def unwind_from_stack(self, sp: int, top: int, max_frames: int = 100) -> List[int]:
        """
        Unwind call stack by analyzing branch instructions on the stack

        This function scans the stack memory from sp to top, looking for
        return addresses that point to BL or BLX instructions. It
        reconstructs the call stack without relying on frame pointers or
        unwind tables.

        Args:
            sp: Stack pointer (bottom of stack to scan)
            top: Top of stack region
            max_frames: Maximum number of frames to collect

        Returns:
            List of return addresses (instruction addresses before the call)
        """
        addresses = []

        while len(addresses) < max_frames and sp < top:
            # Read potential return address from stack
            addr = self.read_word(sp)
            sp += 4
            if addr is None or not self.in_code_region(addr):
                continue

            # ARM Thumb mode: LSB is 1, clear it to get actual address
            # Subtract 2 to get the address of the branch instruction
            addr = (addr & ~1) - 2

            # Verify the address is still in code region
            if not self.in_code_region(addr):
                continue

            # Read the instruction at this address
            ins16_low = self.read_word(addr, halfword=True)
            if ins16_low is None:
                continue

            # Check if it's a BLX instruction (16-bit)
            if self.is_thumb_blx_instruction(ins16_low):
                addresses.append(addr)
                continue

            # Check if it might be a BL instruction (32-bit)
            if (ins16_low & IMASK_T_BL_LOW) != IOP_T_BL_LOW:
                continue

            # Need to check high halfword (2 bytes before)
            addr_high = addr - 2
            if not self.in_code_region(addr_high):
                continue

            ins16_high = self.read_word(addr_high, halfword=True)
            if ins16_high is None:
                continue

            # Check if high and low halfwords form a BL instruction
            if self.is_thumb_bl_instruction(ins16_high, ins16_low):
                addresses.append(addr_high)

        return addresses

    def unwind_from_tcb(self, tcb, max_frames: int = 100) -> List[int]:
        """
        Unwind call stack from a TCB (Task Control Block)

        This function extracts stack information from the TCB and performs
        stack-based unwinding. It handles both running and non-running tasks,
        as well as interrupt contexts.

        Args:
            tcb: TCB structure (gdb.Value) or None for current task
            max_frames: Maximum number of frames to collect

        Returns:
            List of return addresses
        """

        if not tcb or int(tcb) == 0:
            return []

        # Get TCB fields
        try:
            # Get stack bounds
            stack_base = int(tcb["stack_base_ptr"])
            stack_size = int(tcb["adj_stack_size"])
            stack_top = stack_base + stack_size
            # Unwind from the saved stack pointer
            if stack_base and stack_top:
                return self.unwind_from_stack(stack_base, stack_top, max_frames)

        except (gdb.error, Exception):
            # Failed to access TCB fields
            pass

        return []


@autocompeletion.complete
class ArmSpBacktrace(gdb.Command):
    """
    Unwind ARM call stack using stack analysis (no unwind tables)

    Usage:
        armspbt                    - Unwind current task
        armspbt <tcb_address>      - Unwind specific task
        armspbt -s <sp> -t <top>   - Unwind stack range
        armspbt -f <max_frames>    - Limit number of frames

    Examples:
        armspbt
        armspbt 0x20001234
        armspbt -s 0x20001000 -t 0x20002000
        armspbt 0x20001234 -f 20
    """

    def __init__(self):
        # Check if arch is arm
        arch = gdb.execute("show architecture", to_string=True)
        if "arm" not in arch.lower():
            return

        super().__init__("armspbt", gdb.COMMAND_USER)

        parser = argparse.ArgumentParser(
            description="ARM stack-based backtrace unwinder"
        )
        parser.add_argument(
            "tcb", nargs="?", type=str, help="TCB address or expression"
        )
        parser.add_argument("-s", "--sp", type=str, help="Stack pointer address")
        parser.add_argument("-t", "--top", type=str, help="Top of stack address")
        parser.add_argument(
            "-f",
            "--frames",
            type=int,
            default=100,
            help="Maximum frames",
        )
        self.parser = parser

    @utils.dont_repeat_decorator
    def invoke(self, args, from_tty):
        try:
            pargs = self.parser.parse_args(gdb.string_to_argv(args))
        except SystemExit:
            return

        unwinder = ArmSpUnwind()

        # Unwind from explicit stack range
        if pargs.sp and pargs.top:
            try:
                sp = int(utils.parse_arg(pargs.sp))
                top = int(utils.parse_arg(pargs.top))
                addresses = unwinder.unwind_from_stack(sp, top, pargs.frames)
            except (gdb.error, Exception) as e:
                gdb.write(f"Error parsing stack range: {e}\n")
                return
        # Unwind from TCB
        elif pargs.tcb:
            try:
                tcb = utils.parse_arg(pargs.tcb)
                addresses = unwinder.unwind_from_tcb(tcb, pargs.frames)
            except (gdb.error, Exception) as e:
                gdb.write(f"Error parsing TCB: {e}\n")
                return
        # Unwind current task
        else:
            thread = gdb.selected_thread()
            if thread is None:
                gdb.write("No thread selected\n")
                return

            pid = utils.get_gdb_thread_pid(thread)
            tcb = utils.get_tcb(pid)
            print(f"Unwinding current task TCB: {hex(tcb)}, {utils.get_task_name(tcb)}")
            print(
                f"Stack: {tcb.stack_base_ptr} - {tcb.stack_base_ptr + tcb.adj_stack_size}"
            )

            addresses = unwinder.unwind_from_tcb(tcb, pargs.frames)

        if not addresses:
            gdb.write("No backtrace found\n")
            return

        # Format and display backtrace
        backtrace = Backtrace(addresses, break_null=False)
        gdb.write(str(backtrace))
        gdb.write(f"\nTotal frames: {len(addresses)}\n")



############################################################################
# tools/pynuttx/nxtrace/perfetto_trace.py
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import logging
from enum import IntEnum

from . import perfetto_trace_pb2 as pb2

try:
    from google.protobuf.message_factory import GetMessageClass
except ImportError:
    print("Please execute the following command to install dependencies:")
    print("pip install protobuf==4.25.3")
    exit(1)


logger = logging.getLogger(__name__)
logger.setLevel(logging.WARN)


class LogPriority:
    UNSPECIFIED = pb2.AndroidLogPriority.PRIO_UNSPECIFIED
    UNUSED = pb2.AndroidLogPriority.PRIO_UNUSED
    VERBOSE = pb2.AndroidLogPriority.PRIO_VERBOSE
    DEBUG = pb2.AndroidLogPriority.PRIO_DEBUG
    INFO = pb2.AndroidLogPriority.PRIO_INFO
    WARN = pb2.AndroidLogPriority.PRIO_WARN
    ERROR = pb2.AndroidLogPriority.PRIO_ERROR
    FATAL = pb2.AndroidLogPriority.PRIO_FATAL


class TaskState(IntEnum):
    RUNNING = 0
    INTERRUPTIBLE = 1
    UNINTERRUPTIBLE = 2
    STOPPED = 4
    TRACED = 8
    DEAD = 16
    ZOMBIE = 32
    PARKED = 64
    INVALID = 128


class TaskInfo:
    def __init__(self, comm, pid, prio, tid=None):
        self.pid = pid
        self.comm = comm
        self.prio = prio
        self.tid = tid if tid is not None else pid


class TraceHead:
    def __init__(self, ts, pid, cpu, tid=None):
        self.ts = ts
        self.pid = pid
        self.cpu = cpu
        self.tid = tid if tid is not None else pid


class PerfettoTraceBadArgsError(ValueError):
    pass


class TraceClassFactory:
    def __new__(cls, message_name):
        message_descriptor = cls._get_message_descriptor(message_name)
        return GetMessageClass(message_descriptor)

    @classmethod
    def _get_message_descriptor(cls, message_name):
        """
        Recursively find the descriptor for nested messages.
        """
        parts = message_name.split(".")
        descriptor = pb2.DESCRIPTOR

        for part in parts:
            # Handle FileDescriptor
            if hasattr(descriptor, "message_types_by_name"):
                if part not in descriptor.message_types_by_name:
                    raise ValueError(f"Message {message_name} not found in descriptor.")
                descriptor = descriptor.message_types_by_name[part]
            # Handle MessageDescriptor
            elif hasattr(descriptor, "nested_types_by_name"):
                if part not in descriptor.nested_types_by_name:
                    raise ValueError(f"Message {message_name} not found in descriptor.")
                descriptor = descriptor.nested_types_by_name[part]
            else:
                raise ValueError(f"Invalid descriptor type for {message_name}")
        return descriptor


class TraceInstanceFactory:
    def __new__(cls, message_name, **kwargs):
        # Get the message class
        message_class = TraceClassFactory(message_name)
        instance = message_class()
        for field, value in kwargs.items():
            setattr(instance, field, value)
        return instance


class PerfettoTrace:
    DEFAULT_TRUSTED_PACKET_SEQUENCE_ID = 0

    SLICE_BEGIN = pb2.TrackEvent.Type.TYPE_SLICE_BEGIN
    SLICE_END = pb2.TrackEvent.Type.TYPE_SLICE_END
    INSTANT = pb2.TrackEvent.Type.TYPE_INSTANT
    COUNTER = pb2.TrackEvent.Type.TYPE_COUNTER

    def __init__(self, filename: str):
        """Create a trace"""
        self.flush_threshold = 10000

        self.trace = pb2.Trace()
        self.file = open(filename, "wb")

        self.current_uuid = 1
        self.counter_track_cache = {}

    def fill_debug_annotation_value(self, debug_annotation, value):
        if isinstance(value, dict):
            for key in value:
                dict_entry = pb2.DebugAnnotation()
                dict_entry.name = key
                self.fill_debug_annotation_value(dict_entry, value[key])
                debug_annotation.dict_entries.append(dict_entry)
        elif isinstance(value, list):
            for item in value:
                array_value = pb2.DebugAnnotation()
                self.fill_debug_annotation_value(array_value, item)
                debug_annotation.array_values.append(array_value)
        elif isinstance(value, bool):
            debug_annotation.bool_value = value
        elif isinstance(value, int):
            debug_annotation.int_value = value
        elif isinstance(value, float):
            debug_annotation.double_value = value
        elif isinstance(value, str):
            debug_annotation.string_value = value
        elif value is None:
            debug_annotation.string_value = "None"
        else:
            raise PerfettoTraceBadArgsError(
                f"cannot convert value {value} of type {type(value)} to DebugAnnotation value"
            )

    def next_uuid(self):
        self.current_uuid += 1
        return self.current_uuid

    def init(self):
        # Initialize clock snapshots
        clocks_class = TraceClassFactory("ClockSnapshot.Clock")
        clocks = [clocks_class(clock_id=id, timestamp=0) for id in range(1, 7)]

        clock_snapshot = TraceInstanceFactory("ClockSnapshot")
        clock_snapshot.primary_trace_clock = pb2.BUILTIN_CLOCK_BOOTTIME
        clock_snapshot.clocks.extend(clocks)

        pkt = TraceInstanceFactory("TracePacket", trusted_packet_sequence_id=1)
        pkt.clock_snapshot.CopyFrom(clock_snapshot)
        self.trace.packet.append(pkt)

        # Configure trace packets
        pkt = self.trace.packet.add(trusted_packet_sequence_id=1)
        pkt.trace_config.buffers.add().size_kb = 1024
        pkt.trace_config.data_sources.add().config.name = "track_event"

        pkt = self.trace.packet.add(trusted_packet_sequence_id=2, sequence_flags=1)
        pkt.trace_packet_defaults.track_event_defaults.track_uuid = 1
        pkt.trace_packet_defaults.timestamp_clock_id = 1

        pkt = self.trace.packet.add()
        pkt.trusted_packet_sequence_id = 1
        pkt.trace_config.buffers.add().size_kb = 1024
        pkt.trace_config.data_sources.add().config.name = "track_event"

        pkt = self.trace.packet.add()
        pkt.trusted_packet_sequence_id = 2
        pkt.trace_packet_defaults.track_event_defaults.track_uuid = 1
        pkt.trace_packet_defaults.timestamp_clock_id = 1
        pkt.sequence_flags = 1

    def flush(self):
        """Flush trace. This creates a perfetto trace packet and writes to disk."""
        self.file.write(self.trace.SerializeToString())
        self.file.flush()
        self.trace = pb2.Trace()

    def __del__(self):
        self.flush()
        self.file.close()

    # Ftrace common events API

    def log(self, ts, tag, pid, tid=None, prio=LogPriority.INFO, msg="", args=None):
        tid = tid if tid is not None else pid
        pkt = self.trace.packet.add()

        log_event = TraceInstanceFactory(
            "AndroidLogPacket.LogEvent",
            timestamp=ts,
            log_id=pb2.AndroidLogId.LID_DEFAULT,
            pid=pid,
            tid=tid,
            uid=0,
            tag=tag,
            prio=prio,
            message=msg,
        )
        pkt.android_log.events.append(log_event)
        logger.debug(pkt)

    def ftrace_event(self, head: TraceHead, event):
        pkt = self.trace.packet.add()
        ftrace_events = TraceInstanceFactory("FtraceEventBundle", cpu=head.cpu)
        events = TraceInstanceFactory("FtraceEvent", pid=head.pid, timestamp=head.ts)

        event_descriptor = event.DESCRIPTOR

        for oneof in events.DESCRIPTOR.oneofs:
            for field in oneof.fields:
                if field.message_type == event_descriptor:
                    getattr(events, field.name).CopyFrom(event)
                    break

        ftrace_events.event.append(events)
        pkt.ftrace_events.CopyFrom(ftrace_events)
        logger.debug(pkt)
        return pkt

    # Ftrace events API

    def sched_wakeup_new(self, head: TraceHead, task: TaskInfo):
        wakeup = TraceInstanceFactory(
            "SchedWakeupNewFtraceEvent",
            pid=task.pid,
            comm=task.comm,
            prio=task.prio,
            target_cpu=head.cpu,
        )
        return self.ftrace_event(head, wakeup)

    def sched_waking(self, head: TraceHead, task: TaskInfo):
        waking = TraceInstanceFactory(
            "SchedWakingFtraceEvent",
            pid=task.pid,
            comm=task.comm,
            prio=task.prio,
            target_cpu=head.cpu,
        )
        return self.ftrace_event(head, waking)

    def sched_wakeup(self, head: TraceHead, task: TaskInfo):
        wakeup = TraceInstanceFactory(
            "SchedWakeupFtraceEvent",
            pid=task.pid,
            comm=task.comm,
            prio=task.prio,
            target_cpu=head.cpu,
        )
        return self.ftrace_event(head, wakeup)

    def sched_switch(
        self,
        head: TraceHead,
        prev_state=0,
        task_prev: TaskInfo = None,
        task_next: TaskInfo = None,
    ):
        switch = pb2.SchedSwitchFtraceEvent(
            prev_comm=task_prev.comm,
            prev_pid=task_prev.pid,
            prev_prio=task_prev.prio,
            prev_state=prev_state,
            next_comm=task_next.comm,
            next_pid=task_next.pid,
            next_prio=task_next.prio,
        )

        return self.ftrace_event(head, switch)

    def irq_entry(self, head: TraceHead, irq: int, name: str, handler: int):
        irq_entry = TraceInstanceFactory(
            "IrqHandlerEntryFtraceEvent", irq=irq, name=name, handler=handler
        )
        return self.ftrace_event(head, irq_entry)

    def irq_exit(self, head: TraceHead, irq, ret=0):
        irq_exit = TraceInstanceFactory("IrqHandlerExitFtraceEvent", irq=irq, ret=ret)
        return self.ftrace_event(head, irq_exit)

    def print(self, head: TraceHead, buf=None):
        print_event = TraceInstanceFactory("PrintFtraceEvent", buf=buf)
        return self.ftrace_event(head, print_event)

    # Atrace events API

    def atrace_begin(self, head: TraceHead, msg):
        return self.print(head, f"B|{head.pid}|{msg}")

    def atrace_end(self, head: TraceHead, msg=""):
        return self.print(head, f"E|{head.pid}|{msg}" if msg else "")

    def atrace_async_begin(self, head: TraceHead, msg, cookie):
        return self.print(head, f"S|{head.pid}|{msg}|{cookie}")

    def atrace_async_end(self, head: TraceHead, msg, cookie):
        return self.print(head, f"F|{head.pid}|{msg}|{cookie}")

    def atrace_async_for_track_begin(self, head: TraceHead, track_name, msg, cookie):
        return self.print(head, f"G|{head.pid}|{track_name}|{msg}|{cookie}")

    def atrace_async_for_track_end(self, head: TraceHead, track_name, msg, cookie):
        return self.print(head, f"H|{head.pid}|{track_name}|{msg}|{cookie}")

    def atrace_instant(self, head: TraceHead, msg):
        return self.print(head, f"I|{head.pid}|{msg}")

    def atrace_instant_for_track(self, head: TraceHead, track_name, msg):
        return self.print(head, f"N|{head.pid}|{track_name}|{msg}")

    def atrace_int(self, head: TraceHead, msg, value):
        return self.print(head, f"C|{head.pid}|{msg}|{value}")

    def add_process(self, uuid, pid, name):
        self.trace.packet.append(
            pb2.TracePacket(
                track_descriptor=pb2.TrackDescriptor(
                    uuid=uuid, process=pb2.ProcessDescriptor(pid=pid, process_name=name)
                ),
                trusted_packet_sequence_id=self.DEFAULT_TRUSTED_PACKET_SEQUENCE_ID,
            )
        )

    def add_thread(self, uuid, parent_uuid, tid, pid, name):
        self.trace.packet.append(
            pb2.TracePacket(
                track_descriptor=pb2.TrackDescriptor(
                    uuid=uuid,
                    parent_uuid=parent_uuid,
                    thread=pb2.ThreadDescriptor(pid=pid, tid=tid, thread_name=name),
                ),
                trusted_packet_sequence_id=self.DEFAULT_TRUSTED_PACKET_SEQUENCE_ID,
            )
        )

    def add_counter_track(
        self, uuid, parent_uuid, name, unit="UNIT_COUNT", categories=None
    ):
        counter_descriptor = pb2.CounterDescriptor(
            unit=getattr(
                pb2.CounterDescriptor.Unit, unit, pb2.CounterDescriptor.Unit.UNIT_COUNT
            )
        )

        if categories:
            counter_descriptor.categories.extend(categories)

        track_descriptor = pb2.TrackDescriptor(
            uuid=uuid, parent_uuid=parent_uuid, name=name, counter=counter_descriptor
        )

        self.trace.packet.append(
            pb2.TracePacket(
                track_descriptor=track_descriptor,
                trusted_packet_sequence_id=self.DEFAULT_TRUSTED_PACKET_SEQUENCE_ID,
            )
        )

    def trace_event(
        self, uuid, ts, type, name=None, value: int = None, flow_ids=[], args={}
    ):
        pkt = self.trace.packet.add()
        pkt.timestamp = ts
        pkt.track_event.type = type
        pkt.track_event.track_uuid = uuid
        if value is not None:
            pkt.track_event.counter_value = value

        pkt.trusted_packet_sequence_id = self.DEFAULT_TRUSTED_PACKET_SEQUENCE_ID

        if args:
            if not isinstance(args, dict):
                raise PerfettoTraceBadArgsError("slice args is not dict")
            annotations = []

            for key in args:
                debug_annotation = pb2.DebugAnnotation()
                debug_annotation.name = key
                self.fill_debug_annotation_value(debug_annotation, args[key])

                annotations.append(debug_annotation)

            pkt.track_event.debug_annotations.extend(annotations)

        if name:
            pkt.track_event.name = name
        if flow_ids:
            for flow_id in flow_ids:
                pkt.track_event.flow_ids.append(flow_id)

    def trace_slice(self, uuid, start, end, name, flow_ids=[], args={}):
        self.trace_event(
            uuid, start, self.SLICE_BEGIN, name, flow_ids=flow_ids, args=args
        )
        self.trace_event(uuid, end, self.SLICE_END)

    def trace_begin(self, uuid, start, name, flow_ids=[], args={}):
        self.trace_event(
            uuid, start, self.SLICE_BEGIN, name, flow_ids=flow_ids, args=args
        )

    def trace_end(self, uuid, end, name="", flow_ids=[], args={}):
        self.trace_event(uuid, end, self.SLICE_END, name, flow_ids=flow_ids, args=args)

    def trace_counter(self, parent_uuid, ts, name, value: int, flow_ids=[], args={}):
        cache_key = (parent_uuid, name)

        # Check if the counter track already exists
        if cache_key not in self.counter_track_cache:
            counter_uuid = self.next_uuid()
            self.add_counter_track(
                uuid=counter_uuid, parent_uuid=parent_uuid, name=name, unit="UNIT_COUNT"
            )

            self.counter_track_cache[cache_key] = counter_uuid
            logger.debug(
                f"Created counter track: {name} for parent PID {parent_uuid}, UUID: {counter_uuid}"
            )

        # Get the cached counter track UUID
        counter_uuid = self.counter_track_cache[cache_key]

        # Use the counter track UUID to record the event
        return self.trace_event(
            counter_uuid, ts, self.COUNTER, name, value, flow_ids, args
        )



############################################################################
# tools/pynuttx/gdbrpc/cli.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import logging

from gdbrpc.client import Client
from gdbrpc.utils import ShellExec


class ClientCLI(Client):
    def __init__(self, host, port):
        super().__init__(host, port, log_level=logging.WARNING)

    def _show_command_help(self):
        print("Welcome to the GDB Remote Protocol Client")
        print("Type `exit` or `quit` to disconnect.")
        print("Type `help` to show this help message.")
        print("If you need `interrupt` command to stop the target, use Ctrl+C.")

    def _loop(self):
        print("gdb> ", end="", flush=True)
        while True:
            try:
                command = input().strip()
                if not command:
                    continue
                if command.lower() in ("exit", "quit"):
                    self.disconnect()
                    break
                if command.lower() == "help":
                    self._show_command_help()
                    print("gdb> ", end="", flush=True)
                    continue
                print(self.call(ShellExec(command)))
                print("gdb> ", end="", flush=True)
            except KeyboardInterrupt:
                print("")
                self.call(ShellExec("interrupt"))
                print("gdb> ", end="", flush=True)
            except EOFError:
                self.disconnect()

    def start(self):
        assert self.connect()
        self._show_command_help()
        self._loop()



############################################################################
# tools/pynuttx/nxgdb/net.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

from enum import Enum

import gdb

from . import utils
from .lists import NxDQueue, NxSQueue

socket = utils.import_check(
    "socket", errmsg="No socket module found, please try gdb-multiarch instead.\n"
)

CONFIG_NET = utils.lookup_type("struct socket") is not None
if CONFIG_NET:
    CONFIG_NET_IPv4 = utils.has_field("struct net_driver_s", "d_ipaddr")
    CONFIG_NET_IPv6 = utils.has_field("struct net_driver_s", "d_ipv6draddr")
else:
    CONFIG_NET_IPv4 = False
    CONFIG_NET_IPv6 = False

# See include/sys/socket.h
PF_INET = 2
PF_INET6 = 10

AF_INET = PF_INET
AF_INET6 = PF_INET6


def ntohs(val):
    """Convert a 16-bit value from network byte order to host byte order"""

    if utils.get_target_endianness() == utils.BIG_ENDIAN:
        return val
    return utils.swap16(val)


def get_ip_port(conn):
    """Get the IP address and port of a network connection"""

    domain = utils.get_field(conn, "domain", AF_INET if CONFIG_NET_IPv4 else AF_INET6)
    ip_binding = conn["u"]["ipv4" if domain == AF_INET else "ipv6"]
    lport = ntohs(conn["lport"])
    rport = ntohs(conn["rport"])
    laddr = inet_ntop(domain, ip_binding["laddr"])
    raddr = inet_ntop(domain, ip_binding["raddr"])
    return laddr, lport, raddr, rport


def inet_ntop(domain, addr):
    """Convert a network address to a string"""

    addr_len = 16 if domain == AF_INET6 else 4
    if socket:
        return socket.inet_ntop(domain, utils.get_bytes(addr, addr_len))
    else:
        separator = "." if domain == AF_INET else ""
        fmt = "%d" if domain == AF_INET else "%02x"
        return separator.join([fmt % byte for byte in utils.get_bytes(addr, addr_len)])


def socket_for_each_entry(proto):
    """Iterate over a dq of socket structs, usage:
    for conn in socket_for_each_entry("icmp"):
        readahead = conn["readahead"]
    """

    active_connections = utils.parse_and_eval("g_active_%s_connections" % proto)
    for node in NxDQueue(active_connections, "struct socket_conn_s", "node"):
        # udp_conn_s::socket_conn_s sconn
        yield utils.container_of(
            node,
            "struct %s_conn_s" % proto,
            "sconn",
        )


def wrbuffer_inqueue_size(queue=None, protocol="tcp"):
    """Calculate the total size of all iob in the write queue of a udp connection"""

    if not queue:
        return 0

    type = "struct %s_wrbuffer_s" % protocol
    node = "wb_node"
    return sum(entry["wb_iob"]["io_pktlen"] for entry in NxSQueue(queue, type, node))


def tcp_ofoseg_bufsize(conn):
    """Calculate the pending size of out-of-order buffer of a tcp connection"""

    total = 0
    CONFIG_NET_TCP_OUT_OF_ORDER = utils.has_field("struct tcp_conn_s", "ofosegs")
    if CONFIG_NET_TCP_OUT_OF_ORDER:
        total = sum(
            seg["data"]["io_pktlen"]
            for seg in utils.ArrayIterator(conn["ofosegs"], conn["nofosegs"])
        )
    return total


class NetStats(gdb.Command):
    """Network statistics
    Usage: netstats [iob|pkt|tcp|udp|all]

    Examples: netstats - Show all stats
              netstats all - Show all stats
              netstats iob - Show IOB stats
              netstats tcp udp - Show both TCP and UDP stats
    """

    def __init__(self):
        if CONFIG_NET:
            super().__init__("netstats", gdb.COMMAND_USER)

    def iob_stats(self):
        try:
            size = utils.get_symbol_value("CONFIG_IOB_BUFSIZE")
            ntotal = utils.get_symbol_value("CONFIG_IOB_NBUFFERS")

            nfree = utils.parse_and_eval("g_iob_sem")["val"]["semcount"]
            nwait, nfree = (0, nfree) if nfree >= 0 else (-nfree, 0)

            nthrottle = (
                utils.parse_and_eval("g_throttle_sem")["val"]["semcount"]
                if utils.get_symbol_value("CONFIG_IOB_THROTTLE") > 0
                else 0
            )

            gdb.write(
                "IOB: %10s%10s%10s%10s%10s\n"
                % ("size", "ntotal", "nfree", "nwait", "nthrottle")
            )
            gdb.write(
                "     %10d%10d%10d%10d%10d\n" % (size, ntotal, nfree, nwait, nthrottle)
            )
        except gdb.error as e:
            gdb.write("Failed to get IOB stats: %s\n" % e)

    def pkt_stats(self):
        try:
            netstats = utils.parse_and_eval("g_netstats")
            gdb.write(
                "Packets:%7s%7s%7s%7s%7s%7s\n"
                % ("IPv4", "IPv6", "TCP", "UDP", "ICMP", "ICMPv6")
            )

            def stats_line(title, member):
                gdb.write("%-8s" % title)
                for proto in ("ipv4", "ipv6", "tcp", "udp", "icmp", "icmpv6"):
                    gdb.write(
                        "%7s"
                        % utils.get_field(utils.get_field(netstats, proto), member, "-")
                    )
                gdb.write("\n")

            stats_line("Received", "recv")
            stats_line("Dropped", "drop")
            stats_line("  VHL", "vhlerr")
            stats_line("  Frag", "fragerr")
            stats_line("  Chksum", "chkerr")
            stats_line("  Type", "typeerr")
            stats_line("  Proto", "protoerr")
            # TODO: Maybe print TCP's ackerr, rst, syndrop, synrst here
            stats_line("Sent", "sent")
            stats_line("  Rexmit", "rexmit")

        except gdb.error as e:
            gdb.write("Failed to get Net Stats: %s\n" % e)

    def tcp_stats(self):
        try:
            gdb.write(
                "TCP Conn: %3s %3s %3s %3s %4s %3s"
                % ("st", "flg", "ref", "tmr", "uack", "nrt")
            )
            gdb.write(
                " %11s %11s+%-5s %21s %21s\n"
                % ("txbuf", "rxbuf", "ofo", "local_address", "remote_address")
            )
            for idx, conn in enumerate(socket_for_each_entry("tcp")):
                state = conn["tcpstateflags"]
                flags = conn["sconn"]["s_flags"]
                ref = conn["crefs"]
                timer = conn["timer"]
                unacked = conn["tx_unacked"]
                nrtx = conn["nrtx"]

                txbuf = utils.get_field(conn, "snd_bufs", -1)
                rxbuf = utils.get_field(conn, "rcv_bufs", -1)
                txsz = wrbuffer_inqueue_size(
                    utils.get_field(conn, "unacked_q"), "tcp"
                ) + wrbuffer_inqueue_size(utils.get_field(conn, "write_q"), "tcp")
                rxsz = conn["readahead"]["io_pktlen"] if conn["readahead"] else 0
                ofosz = tcp_ofoseg_bufsize(conn)
                laddr, lport, raddr, rport = get_ip_port(conn)

                gdb.write(
                    "%-4d      %3x %3x %3d %3d %4d %3d"
                    % (idx, state, flags, ref, timer, unacked, nrtx)
                )
                gdb.write(
                    " %5d/%-5d %5d/%-5d+%-5d %15s:%-5d %15s:%-5d\n"
                    % (txsz, txbuf, rxsz, rxbuf, ofosz, laddr, lport, raddr, rport)
                )
        except gdb.error as e:
            gdb.write("Failed to get TCP stats: %s\n" % e)

    def udp_stats(self):
        try:
            gdb.write(
                "UDP Conn: %4s %11s %11s %21s %21s\n"
                % ("flg", "txbuf", "rxbuf", "local_address", "remote_address")
            )
            for idx, conn in enumerate(socket_for_each_entry("udp")):
                flags = conn["sconn"]["s_flags"]
                txbuf = utils.get_field(conn, "sndbufs", -1)
                rxbuf = utils.get_field(conn, "rcvbufs", -1)
                txsz = wrbuffer_inqueue_size(utils.get_field(conn, "write_q"), "udp")
                rxsz = conn["readahead"]["io_pktlen"] if conn["readahead"] else 0
                laddr, lport, raddr, rport = get_ip_port(conn)

                gdb.write(
                    "%-4d      %4x %5d/%-5d %5d/%-5d %15s:%-5d %15s:%-5d\n"
                    % (idx, flags, txsz, txbuf, rxsz, rxbuf, laddr, lport, raddr, rport)
                )
        except gdb.error as e:
            gdb.write("Failed to get UDP stats: %s\n" % e)

    @utils.dont_repeat_decorator
    def invoke(self, args, from_tty):
        # Parse the arguments in a simple way
        if not args or "all" in args:
            args = "iob pkt tcp udp"

        # Call the corresponding function
        if utils.get_symbol_value("CONFIG_MM_IOB") and "iob" in args:
            self.iob_stats()
            gdb.write("\n")
        if utils.get_symbol_value("CONFIG_NET_STATISTICS") and "pkt" in args:
            self.pkt_stats()
            gdb.write("\n")
        if utils.get_symbol_value("CONFIG_NET_TCP") and "tcp" in args:
            self.tcp_stats()
            gdb.write("\n")
        if utils.get_symbol_value("CONFIG_NET_UDP") and "udp" in args:
            self.udp_stats()
            gdb.write("\n")


class NetCheckResult(Enum):
    PASS = 0
    WARN = 1
    FAILED = 2

    def __lt__(self, other):
        return self.value < other.value


class NetCheck(gdb.Command):
    """Network check"""

    def __init__(self):
        if CONFIG_NET:
            super().__init__("netcheck", gdb.COMMAND_USER)

    def diagnose(self, *args, **kwargs):
        result, message = NetCheckResult.PASS, []

        if utils.get_symbol_value("CONFIG_MM_IOB"):
            ret, msg = self.check_iob()
            result = max(result, ret)
            message.extend(msg)

        return {
            "title": "Netcheck Report",
            "summary": "Net status check",
            "result": "pass" if result else "fail",
            "category": utils.DiagnoseCategory.connectivity,
            "command": "netcheck",
            "data": message,
        }

    def check_iob(self):
        result = NetCheckResult.PASS
        message = []
        try:
            nfree = utils.parse_and_eval("g_iob_sem")["val"]["semcount"]
            nthrottle = (
                utils.parse_and_eval("g_throttle_sem")["val"]["semcount"]
                if utils.get_symbol_value("CONFIG_IOB_THROTTLE") > 0
                else 0
            )

            if nfree < 0 or nthrottle < 0:
                result = max(result, NetCheckResult.WARN)
                message.append(
                    "[WARNING] IOB used up: free %d throttle %d" % (nfree, nthrottle)
                )

        except gdb.error as e:
            result = max(result, NetCheckResult.FAILED)
            message.append("[FAILED] Failed to check IOB: %s" % e)

        finally:
            return result, message

    @utils.dont_repeat_decorator
    def invoke(self, args, from_tty):
        if utils.get_symbol_value("CONFIG_MM_IOB"):
            result, message = self.check_iob()
            gdb.write("IOB check: %s\n" % result.name)
            gdb.write("\n".join(message))
